@@ -62,6 +62,7 @@ static GMainLoop *main_loop = NULL;
 
 static int services = 0;
 static gboolean tty_needs_reinit = FALSE;
+static gboolean tty_open_allowed = TRUE;
 static int signal_pipe[2];
 
 #define TTY_RX_MTU 65535
@@ -72,11 +73,15 @@ int tty_init(int services, const gchar *root_path,
 {
 	struct server *server;
 	struct termios options;
-	int fd, ret;
+	int fd, err, arg;
 	glong flags;
+	GIOChannel *io = NULL;
 
 	tty_needs_reinit = TRUE;
 
+	if (!tty_open_allowed)
+		return -EACCES;
+
 	fd = open(devnode, O_RDWR | O_NOCTTY);
 	if (fd < 0)
 		return fd;
@@ -88,6 +93,18 @@ int tty_init(int services, const gchar *root_path,
 	cfmakeraw(&options);
 	tcsetattr(fd, TCSANOW, &options);
 
+	arg = fcntl(fd, F_GETFL);
+	if (arg < 0) {
+		err = -errno;
+		goto failed;
+	}
+
+	arg |= O_NONBLOCK;
+	if (fcntl(fd, F_SETFL, arg) < 0) {
+		err = -errno;
+		goto failed;
+	}
+
 	server = g_new0(struct server, 1);
 	server->services = services;
 	server->folder = g_strdup(root_path);
@@ -97,14 +114,28 @@ int tty_init(int services, const gchar *root_path,
 	server->rx_mtu = TTY_RX_MTU;
 	server->tx_mtu = TTY_TX_MTU;
 
-	ret = obex_session_start(fd, server);
-	if (ret < 0) {
-		server_free(server);
-		close(fd);
-	} else
-		tty_needs_reinit = FALSE;
+	io = g_io_channel_unix_new(fd);
+	g_io_channel_set_close_on_unref(io, TRUE);
 
-	return ret;
+	err = obex_session_start(io, server);
+	g_io_channel_unref(io);
+
+	if (err < 0) {
+		server_free(server);
+		goto failed;
+	}
+
+	tty_needs_reinit = FALSE;
+
+	debug("Successfully opened %s", devnode);
+
+	return 0;
+
+failed:
+	error("tty_init(): %s (%d)", strerror(-err), -err);
+	if (io == NULL)
+		close(fd);
+	return err;
 }
 
 void tty_closed(void)
@@ -114,9 +145,15 @@ void tty_closed(void)
 
 static void sig_term(int sig)
 {
+	info("Terminating due to signal %d", sig);
 	g_main_loop_quit(main_loop);
 }
 
+static void sig_debug(int sig)
+{
+	toggle_debug();
+}
+
 static gboolean option_detach = TRUE;
 static gboolean option_debug = FALSE;
 
@@ -142,7 +179,7 @@ static GOptionEntry options[] = {
 	{ "symlinks", 'l', 0, G_OPTION_ARG_NONE, &option_symlinks,
 				"Enable symlinks on root folder" },
 	{ "capability", 'c', 0, G_OPTION_ARG_STRING, &option_capability,
-				"Sepcify capability file", "FILE" },
+				"Specify capability file", "FILE" },
 	{ "tty", 't', 0, G_OPTION_ARG_STRING, &option_devnode,
 				"Specify the TTY device", "DEVICE" },
 	{ "auto-accept", 'a', 0, G_OPTION_ARG_NONE, &option_autoaccept,
@@ -158,7 +195,7 @@ static GOptionEntry options[] = {
 	{ NULL },
 };
 
-static void sig_usr1(int sig)
+static void sig_tty(int sig)
 {
 	if (write(signal_pipe[1], &sig, sizeof(sig)) != sizeof(sig))
 		error("unable to write to signal pipe");
@@ -170,13 +207,27 @@ static gboolean handle_signal(GIOChannel *io, GIOCondition cond,
 	int sig, fd = g_io_channel_unix_get_fd(io);
 
 	if (read(fd, &sig, sizeof(sig)) != sizeof(sig)) {
-		error("handle_sigusr1: unable to read signal from pipe");
+		error("handle_signal: unable to read signal from pipe");
 		return TRUE;
 	}
 
-	if (sig == SIGUSR1 && tty_needs_reinit)
-		tty_init(services, option_root, option_capability,
-				option_devnode);
+	switch (sig) {
+	case SIGUSR1:
+		debug("SIGUSR1");
+		tty_open_allowed = TRUE;
+		if (tty_needs_reinit)
+			tty_init(services, option_root, option_capability,
+							option_devnode);
+		break;
+	case SIGHUP:
+		debug("SIGHUP");
+		tty_open_allowed = FALSE;
+		obex_tty_session_stop();
+		break;
+	default:
+		error("handle_signal: got unexpected signal %d", sig);
+		break;
+	}
 
 	return TRUE;
 }
@@ -194,13 +245,12 @@ static int devnode_setup(void)
 	g_io_channel_unref(pipe_io);
 
 	memset(&sa, 0, sizeof(sa));
-	sa.sa_handler = sig_usr1;
+	sa.sa_handler = sig_tty;
 	sigaction(SIGUSR1, &sa, NULL);
+	sigaction(SIGHUP, &sa, NULL);
 
-	if (option_pcsuite) {
-		tty_needs_reinit = TRUE;
-		return 0;
-	}
+	if (option_pcsuite)
+		tty_open_allowed = FALSE;
 
 	return tty_init(services, option_root, option_capability,
 			option_devnode);
@@ -312,6 +362,9 @@ int main(int argc, char *argv[])
 	sigaction(SIGINT, &sa, NULL);
 	sigaction(SIGTERM, &sa, NULL);
 
+	sa.sa_handler = sig_debug;
+	sigaction(SIGUSR2, &sa, NULL);
+
 	g_main_loop_run(main_loop);
 
 	bluetooth_exit();
